/*
 * Copyright (c) 2010-2015 Yuichi Watanabe
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <common/list.h>
#include <core/assert.h>
#include <core/initfunc.h>
#include <core/param.h>
#include <core/printf.h>
#include <core/spinlock.h>
#include <core/mm.h>
#include <core/mmio.h>
#include <core/rm.h>
#include <core/string.h>
#include <core/vm.h>
#include <core/sleep.h>
#include <core/vm_config.h>
#include <core/vmmerr.h>
#include "pci.h"
#include "pci_internal.h"
#include "vmrm.h"

/* #define PCIASSIGN_DEBUG */
#ifdef PCIASSIGN_DEBUG
#define PCIASSIGN_DBG(...)						\
	do {							\
		printf("PCIASSIGN: " __VA_ARGS__);		\
	} while (0)
#else
#define PCIASSIGN_DBG(...)
#endif

#define PCI_DUMMY_FUNC0_VENDOR_ID 0x0002
#define PCI_DUMMY_FUNC0_DEVICE_ID 0x0001

#define DUMP_ASSIGNMENT
/* #define NO_CLEAR_IOBITMAP */

/* #define NO_SECONDARY_BUS_RESET */
/* #define NO_D3HOT_D0_RESET */
/* #define NO_REINIT_DEVICE */
/* #define REINIT_ALL_REGS */

drvdata_hdl_t		pci_handle;

static size_t
pci_config_size(struct pci_device *dev)
{
	return dev->config_mmio ? PCI_MMCONFIG_SIZE : PCI_LEGACY_CONFIG_SIZE;
}

void
pci_update_saved_config(struct pci_device *dev, pci_off_t offset,
			void *val, uint iosize)
{
	char *config;

	config = (char *)dev->saved_config;
	memcpy(config + offset, val, iosize);
}

static void
pci_save_config(struct pci_device *dev)
{
	int size = pci_config_size(dev);
	int i;

	for (i = 0; i < size; i += 4) {
		dev->saved_config[i / 4] = pci_read_config32(dev, i);
	}
}

static void
pci_save_configs(struct pci_device *dev)
{
	struct pci_device *child;

	pci_save_config(dev);

	LIST2_FOREACH(dev->children, sibling, child) {
		pci_save_configs(child);
	}
}

static void
pci_setup_assigned_device(struct pci_device *dev)
{
	struct pci_device *child;
	struct resource *resource;
	u32 reg_val, mask;
	size_t size;
	vmmerr_t err;

	PCIASSIGN_DBG("setup %s\n", dev->name);

	/*
	 * Disable IRQ for assigned devices.
	 */
	pci_write_config8(dev,
			   PCI_CONFIG_INTERRUPT_LINE,
			   PCI_ASSIGNED_DEVICE_IRQ);

	/*
	 * Disable INTx interrupt.
	 */
	reg_val = pci_read_config16(dev, PCI_CONFIG_COMMAND);
	reg_val |= PCI_CONFIG_COMMAND_INTERRUPT_DISABLE;
	pci_write_config16(dev, PCI_CONFIG_COMMAND, reg_val);

	PCIASSIGN_DBG("disable INTx interrupt %s 0x%x\n",
		      dev->name, reg_val);

	/*
	 * Assign a memory region to ROM.
	 */
	resource = dev->resource + PCI_RESOURCE_EXPANTION_ROM_INDEX;
	mask = pci_get_bar_mask(dev, PCI_CONFIG_EXPANSION_ROM);
	if (resource->type == RESOURCE_TYPE_INVALID &&
	    mask != 0x00000000 && mask != 0xFFFFFFFF) {
		size = ~(mask & PCI_CONFIG_ROM_MEMMASK) + 1;
		err = pci_alloc_resource(dev, resource, size, size,
					 RESOURCE_TYPE_MMIO,
					 PCI_RESOURCE_PREFETCHABLE);
		if (err == VMMERR_SUCCESS) {
			reg_val = resource->start & PCI_CONFIG_ROM_MEMMASK;
			pci_write_config32(dev,
					   PCI_CONFIG_EXPANSION_ROM,
					   reg_val);
		} else {
			printf("Failed to allocate a memory region for ROM. "
			       "pci %s size 0x%lx err 0x%x\n",
			       dev->name, size, err);
		}
	}

	LIST2_FOREACH(dev->children, sibling, child) {
		pci_setup_assigned_device(child);
	}
}

static void
pci_assign_device(struct pci_vm_data *pci_data, struct pci_device *dev,
		  bool is_vm0)
{
#ifndef NO_CLEAR_IOBITMAP
	struct resource *resource;
	int i, j;
#endif

	LIST2_ADD(pci_data->vm_pci_device_list, vm_pci_device_list, dev);

	if (!is_vm0) {
		pci_setup_assigned_device(dev);
	}

	pci_save_configs(dev);

#ifndef NO_CLEAR_IOBITMAP
	/*
	 * Set do_io_pass to clear io bitmap so that vm-exits on io
	 * port access will be reduced.
	 */
	for (i = 0; i < PCI_RESOURCE_NUM; i++) {
		resource = dev->resource + i;
		if (resource->type != RESOURCE_TYPE_IO) {
			continue;
		}
		for (j = resource->start; j <= (int)resource->end; j++) {
			set_iofunc(j, do_iopass_default);
		}
	}
#endif

	if (!is_vm0) {
		dev->assigned = true;
	}
}

static void
pci_dump_device(struct pci_device *dev, int level)
{
	struct pci_device *child;
	char buf[17] = {0};

	memset(buf, ' ', level <= 8 ? level * 2 : 16);
	printf("    %s%s %04x:%04x %02x %02x\n",
	       buf, dev->name,
	       dev->vendor_id, dev->device_id,
	       dev->class_code, dev->header_type);
	LIST2_FOREACH(dev->children, sibling, child) {
		pci_dump_device(child, level + 1);
	}
}

#ifdef DUMP_ASSIGNMENT
void
pci_dump_assignment(void)
{
	struct pci_vm_data *pci_data;
	struct pci_device *dev = NULL;
	static spinlock_t lock = SPINLOCK_INITIALIZER;

	pci_data = vm_get_driver_data(pci_handle);

	spinlock_lock(&lock);
	LIST2_FOREACH(pci_data->vm_pci_device_list, vm_pci_device_list, dev) {
		pci_dump_device(dev, 0);
	}
	spinlock_unlock(&lock);
}
#endif

static void
pci_assign_devices(struct vm_config *vm_config, struct pci_vm_data *pci_data)
{
	struct pci_device *dev = NULL;
	struct pci_config *pci_config;

	while ((dev = pci_next_device_on_root_bus(dev)) != NULL) {
		LIST2_FOREACH(vm_config->pci_list, pci_list, pci_config) {
			if (dev->bus_no == pci_config->bus_no &&
			    dev->devfn == pci_config->devfn) {
				pci_assign_device(pci_data, dev, false);
			}
		}
	}
}

static void
pci_restrict_access(struct pci_device *dev)
{
	u32 port;
	struct resource *resource;
	int i;

	ASSERT(vm_get_id() == 0);

	printf("pci_restrict_access %s\n", dev->name);
	for (i = 0; i < PCI_RESOURCE_NUM; i++) {
		resource = dev->resource + i;
		switch (resource->type) {
		case RESOURCE_TYPE_MMIO:
			mmio_register(resource->start,
				      resource->end - resource->start + 1,
				      mmio_do_nothing, NULL);
			break;
		case RESOURCE_TYPE_IO:
			for (port = resource->start; port <= resource->end;
			     port++) {
				set_iofunc(i, do_io_nothing);
			}
			break;
		}
	}
}

void
pci_assign_devices_to_vm0(struct pci_vm_data *pci_data)
{
	struct pci_device *dev = NULL;

	ASSERT(vm_get_id() == 0);

	/*
	 * This function is called for vm0 after all assignments except
	 * vm0 are completed.
	 */
	while ((dev = pci_next_device_on_root_bus(dev)) != NULL) {
		if (dev->assigned) {
			pci_restrict_access(dev);
			continue;
		}
		pci_assign_device(pci_data, dev, true);
	}
}

static void
pci_preinit(void)
{
	pci_handle = vm_alloc_driver_data(sizeof(struct pci_vm_data));
}

void
pci_assign_all_devices(void)
{
	struct vm_config *vm_config;
	struct pci_vm_data *pci_data;
	char *name;

	LIST2_FOREACH(vm_config_list.head, vm_config_list, vm_config) {
		if (strcmp(vm_config->name, VM0_NAME) == 0) {
				continue;
		}
		name = vm_config->name;
		pci_data = vm_get_driver_data_of_vm(name, pci_handle);
		if (pci_data == NULL) {
			panic("Failed to get pci_data of %s", name);
		}
		LIST2_HEAD_INIT(pci_data->vm_pci_device_list,
				vm_pci_device_list);
		pci_assign_devices(vm_config, pci_data);
	}

	pci_data = vm_get_driver_data_of_vm(VM0_NAME, pci_handle);
	if (pci_data == NULL) {
		panic("Failed to get pci_data of %s", VM0_NAME);
	}
	LIST2_HEAD_INIT(pci_data->vm_pci_device_list,
			vm_pci_device_list);
	pci_assign_devices_to_vm0(pci_data);
}

static void
pci_insert_vmresource(struct pci_device *dev, struct resource *resource)
{
	struct pci_device *parent;
	struct resource *presource;
	vmmerr_t err;
	int j;

	parent = dev->parent;

	if (parent == NULL) {
		switch (resource->type) {
		case RESOURCE_TYPE_MMIO:
			rm_insert_resource(vmrm_mmio_resource(),
					   resource);
			break;
		case RESOURCE_TYPE_IO:
			rm_insert_resource(vmrm_io_resource(),
					   resource);
			break;
		}
	} else {
		for (j = 0; j < PCI_RESOURCE_NUM; j++) {
			presource = parent->vmresource + j;
			if ((presource->data & PCI_RESOURCE_WINDOW)
			    == 0) {
				continue;
			}
			if (resource->type != presource->type) {
				continue;
			}
			if ((presource->data &
			     PCI_RESOURCE_PREFETCHABLE) &&
			    (resource->data &
			     PCI_RESOURCE_PREFETCHABLE) == 0) {
				continue;
			}
			err = rm_insert_resource(presource, resource);
			if (err == VMMERR_SUCCESS) {
				break;
			}
		}
	}
	if (resource->parent == NULL) {
		printf("Fail to insert resource. %s"
		       " %016llx-%016llx type %x data %x\n",
		       dev->name, resource->start,
		       resource->end, resource->type, resource->data);
	}
}

static void
pci_setup_vmresource_of_assigned_dev(struct pci_device *dev)
{
	struct pci_device *child;
	struct resource *resource;
	int i;

	for (i = 0; i < PCI_RESOURCE_NUM; i++) {
		resource = dev->resource + i;
		if (resource->type == RESOURCE_TYPE_INVALID) {
			continue;
		}
		rm_init_resource(dev->vmresource + i, resource->start,
				 resource->end, resource->type,
				 resource->data, resource->name);
		pci_insert_vmresource(dev, resource);
	}

	LIST2_FOREACH(dev->children, sibling, child) {
		pci_setup_vmresource_of_assigned_dev(child);
	}
}

static void
pci_reserve_unused_legacy_ioport(void)
{
	struct resource *resource;
	vmmerr_t err;
	for (;;) {
		resource = alloc(sizeof(struct resource));
		if (resource == NULL) {
			panic("Failed to reserve unused legacy ioport");
		}
		err = rm_alloc_avail_resource(vmrm_io_resource(), resource,
					      0x0, 0xfff, 1,
					      RESOURCE_TYPE_IO, 0, "legacy");
		if (err != VMMERR_SUCCESS) {
			free(resource);
			break;
		}
	}
}

static void
pci_setup_vmresource(void)
{
	struct pci_vm_data *pci_data;
	struct pci_device *dev;
	pci_data = (struct pci_vm_data *)vm_get_driver_data(pci_handle);

	LIST2_FOREACH(pci_data->vm_pci_device_list, vm_pci_device_list, dev) {
		pci_setup_vmresource_of_assigned_dev(dev);
	}
	pci_reserve_unused_legacy_ioport();
}

int
pci_check_assignment(u8 bus_no, u8 devfn)
{
	struct pci_vm_data *pci_data;
	struct pci_device *dev;

	pci_data = (struct pci_vm_data *)vm_get_driver_data(pci_handle);

	LIST2_FOREACH(pci_data->vm_pci_device_list, vm_pci_device_list, dev) {
		if (bus_no == dev->bus_no && devfn == dev->devfn) {
			return PCI_ASSIGNED;
		}
		if (dev->type == PCI_CONFIG_HEADER_TYPE_1 &&
		    bus_no >= dev->sec_bus && bus_no <= dev->sub_bus) {
			return PCI_ASSIGNED;
		}
		if (bus_no == dev->bus_no &&
		    PCI_DEV_NO(devfn) == PCI_DEV_NO(dev->devfn) &&
		    PCI_FUNC_NO(devfn) == 0) {
			return PCI_DUMMY_FUNC0;
		}
	}
	return PCI_NOT_ASSIGNED;
}

struct pci_device *
pci_next_assgined_pci_device(struct pci_device *cur)
{
	struct pci_vm_data *pci_data;
	struct pci_device *dev;

	pci_data = (struct pci_vm_data *)vm_get_driver_data(pci_handle);

	dev = LIST2_NEXT(pci_data->vm_pci_device_list,
			 vm_pci_device_list, cur);
	return dev;
}

int
pci_dummy_func0(pci_off_t offset, pci_off_t size, bool wr, void *data)
{
	static char header[0x10 + 3] = {PCI_DUMMY_FUNC0_VENDOR_ID,
					PCI_DUMMY_FUNC0_VENDOR_ID >> 8,
					PCI_DUMMY_FUNC0_DEVICE_ID,
					PCI_DUMMY_FUNC0_DEVICE_ID >> 8,
					0x0, 0x0, 0x0, 0x0,
					0x0, 0x0, 0x0, 0x0,
					0x0, 0x0,
					PCI_CONFIG_HEADER_TYPE_MULTI_FUNCTION |
					PCI_CONFIG_HEADER_TYPE_0,
					0x0,
					0x0, 0x0, 0x0};

	if (wr) {
		return 1; /* no pass */
	}
	if (offset < sizeof(header) - 3) {
		memcpy(data, header + offset, size);
	} else {
		memset(data, 0, size);
	}
	return 1; /* no pass */
}

int
pci_assigned_mmio_addr(phys_t addr, size_t *len)
{
	struct pci_vm_data *pci_data;
	struct pci_device *dev;
	struct resource *resource;
	int i;

	pci_data = (struct pci_vm_data *)vm_get_driver_data(pci_handle);

	LIST2_FOREACH(pci_data->vm_pci_device_list, vm_pci_device_list, dev) {
		for (i = 0; i < PCI_RESOURCE_NUM; i++) {
			resource = dev->resource + i;
			if (resource->type != RESOURCE_TYPE_MMIO) {
				continue;
			}
			if (resource->start <= addr && resource->end >= addr) {
				if (len) {
					*len = resource->end - addr + 1;
				}
				return 1;
			}
		}
	}
	return 0;
}

int
pci_assigned_ioport(ioport_t port, ioport_t *len)
{
	struct pci_vm_data *pci_data;
	struct pci_device *dev;
	struct resource *resource;
	int i;

	pci_data = (struct pci_vm_data *)vm_get_driver_data(pci_handle);

	LIST2_FOREACH(pci_data->vm_pci_device_list, vm_pci_device_list, dev) {
		for (i = 0; i < PCI_RESOURCE_NUM; i++) {
			resource = dev->resource + i;
			if (resource->type != RESOURCE_TYPE_IO) {
				continue;
			}
			if (resource->start <= port && resource->end >= port) {
				if (len) {
					*len = resource->end - port + 1;
				}
				return 1;
			}
		}
	}
	return 0;
}

static pci_off_t
pci_find_capability(struct pci_device *dev, u8 cap_id)
{
	pci_off_t off;
	u8 id;

	off = pci_read_config8(dev, PCI_CONFIG_CAPABILITY_POINTER);
	for (;;) {
		if (off == 0 || off >= 0xff) {
			return 0;
		}

		id = pci_read_config8(dev, off);
		if (id == cap_id) {
			return off;
		}
		off = pci_read_config8(dev, off + 1);
	}

}

static void
pci_secondary_bus_reset(struct pci_device *dev)
{
#ifndef NO_SECONDARY_BUS_RESET
	u16 reg;
	u16 vendor_id;
	u16 device_id;

	vendor_id = pci_read_config16(dev, PCI_CONFIG_VENDOR_ID);
	device_id = pci_read_config16(dev, PCI_CONFIG_DEVICE_ID);
	if (dev->vendor_id != vendor_id || dev->device_id != device_id) {
		/* A device might be replaced */
		return;
	}

	if (dev->type != PCI_CONFIG_HEADER_TYPE_1) {
		return;
	}
	printf("Execute secondary bus reset on %s\n", dev->name);
	reg = pci_read_config16(dev, PCI_CONFIG_BRIDGE_CONTROL);
	reg |= PCI_CONFIG_SECONDARY_BUS_RESET;
	pci_write_config16(dev, PCI_CONFIG_BRIDGE_CONTROL, reg);

	/* wait 100 ms */
	usleep(100 * 1000);

	reg &= ~PCI_CONFIG_SECONDARY_BUS_RESET;
	pci_write_config16(dev, PCI_CONFIG_BRIDGE_CONTROL, reg);
#endif /* NO_SECONDARY_BUS_RESET */
}

static void
pci_d3d0_reset_device(struct pci_device *dev)
{
	pci_off_t base;
	u16 reg;

	base = pci_find_capability(dev, PCI_PM_CAP_ID);
	if (base == 0) {
		return;
	}

	reg = pci_read_config16(dev, base + PCI_PM_CONTROL_STATUS_REG_OFF);

#ifndef NO_D3HOT_D0_RESET
	if ((reg & PCI_PM_CONTROL_STATUS_REG_NO_SOFT_RESET) == 0) {
		printf("Reset %s\n", dev->name);

		reg = (reg & ~PCI_PM_CONTROL_STATUS_REG_POWER_MASK)
			| PCI_PM_CONTROL_STATUS_REG_POWER_D3HOT;
		pci_write_config16(dev,
				   base + PCI_PM_CONTROL_STATUS_REG_OFF,
				   reg);
	}

	/*wait 10 ms */
	usleep(10 * 1000);
#endif /* NO_D3HOT_D0_RESET */

	reg = (reg & ~PCI_PM_CONTROL_STATUS_REG_POWER_MASK)
		| PCI_PM_CONTROL_STATUS_REG_POWER_D0;
	pci_write_config16(dev, base + PCI_PM_CONTROL_STATUS_REG_OFF, reg);
}

static void
pci_d3d0_reset_devices(struct pci_device *dev)
{
	struct pci_device *child;
	bool has_child = false;
	u16 vendor_id;
	u16 device_id;

	vendor_id = pci_read_config16(dev, PCI_CONFIG_VENDOR_ID);
	device_id = pci_read_config16(dev, PCI_CONFIG_DEVICE_ID);
	if (dev->vendor_id != vendor_id || dev->device_id != device_id) {
		/* A device might be replaced */
		return;
	}

	LIST2_FOREACH(dev->children, sibling, child) {
		pci_d3d0_reset_devices(child);
		has_child = true;
	}

	if (!has_child) {
		pci_d3d0_reset_device(dev);
	}
}

static void
pci_reinit_device(struct pci_device *dev)
{
#ifndef NO_REINIT_DEVICE
#ifdef REINIT_ALL_REGS
	int size = pci_config_size(dev);
	int i;

	printf("Reinit %s\n", dev->name);
	pci_write_config32(dev, PCI_CONFIG_COMMAND, 0x0);
	for (i = PCI_CONFIG_COMMAND + 4; i < size; i += 4) {
		pci_write_config32(dev, i, dev->saved_config[i / 4]);
	}
	pci_write_config32(dev, PCI_CONFIG_COMMAND,
		   dev->saved_config[PCI_CONFIG_COMMAND / 4]);

#else /* REINIT_ALL_REGS */

	pci_off_t base, size;
	int i;
	u16 reg;

	printf("Reinit %s\n", dev->name);
	pci_write_config32(dev, PCI_CONFIG_COMMAND, 0x0);

	for (i = PCI_CONFIG_COMMAND + 4; i < PCI_CONFIG_HEADER_SIZE; i += 4) {
		pci_write_config32(dev, i, dev->saved_config[i / 4]);
	}

	base = pci_find_capability(dev, PCI_PM_CAP_ID);
	if (base != 0) {
		size = PCI_PM_CAP_STRUCTURE_SIZE;
		for (i = base; i < base + size; i += 4) {
			pci_write_config32(dev, i, dev->saved_config[i / 4]);
		}
	}

	base = pci_find_capability(dev, PCI_PCIE_CAP_ID);
	if (base != 0) {
		reg = pci_read_config16(dev, base + PCI_PCIE_CAP_REG_OFF);
		if ((reg & PCI_PCIE_CAP_REG_PCIECV_MASK) >= 2) {
			size = PCI_PCIE_CAP_REV2_STRUCTURE_SIZE;
		} else {
			size = PCI_PCIE_CAP_REV1_STRUCTURE_SIZE;
		}
		for (i = base; i < base + size; i += 4) {
			pci_write_config32(dev, i, dev->saved_config[i / 4]);
		}
	}

	pci_write_config32(dev, PCI_CONFIG_COMMAND,
			   dev->saved_config[PCI_CONFIG_COMMAND / 4]);
#endif /* !REINIT_ALL_REGS */
#endif /* NO_REINIT_DEVICE */
}

static void
pci_reinit_devices(struct pci_device *dev, bool root_bus)
{
	struct pci_device *child;
	pci_off_t base;
	u16 reg;
	u16 vendor_id;
	u16 device_id;

	vendor_id = pci_read_config16(dev, PCI_CONFIG_VENDOR_ID);
	device_id = pci_read_config16(dev, PCI_CONFIG_DEVICE_ID);
	if (dev->vendor_id != vendor_id || dev->device_id != device_id) {
		/* A device might be replaced */
		return;
	}

	if (root_bus &&
	    LIST2_NEXT(dev->children, sibling, (struct pci_device *)NULL)
		== NULL) {
		base = pci_find_capability(dev, PCI_PM_CAP_ID);
		if (base != 0) {
			reg = pci_read_config16(dev,
				base + PCI_PM_CONTROL_STATUS_REG_OFF);
			if ((reg & PCI_PM_CONTROL_STATUS_REG_NO_SOFT_RESET)
			    == 0) {
				/*
				 * VMM re-init device on root bus if
				 * it is reset by d3 to d0 reset.
				 */
				pci_reinit_device(dev);
			}
		}
	} else if (!root_bus) {
		/*
		 * VMM re-init device on non root bus which is reset
		 * by secondary bus reset and d3 to d0 reset.
		 */
		pci_reinit_device(dev);
	}

	if (dev->driver && dev->driver->reinit) {
		dev->driver->reinit(dev);
	}

	LIST2_FOREACH(dev->children, sibling, child) {
		pci_reinit_devices(child, false);
	}

}

static void
pci_reset_assigned_device(void)
{
	struct pci_vm_data *pci_data;
	struct pci_device *dev;

	pci_data = (struct pci_vm_data *)vm_get_driver_data(pci_handle);

	/* Secondary bus reset */
	LIST2_FOREACH(pci_data->vm_pci_device_list, vm_pci_device_list, dev) {
		pci_secondary_bus_reset(dev);
	}

	/* wait 2 sec */
	usleep(2 * 1000 * 1000);

	/* D3hot D0 reset */
	LIST2_FOREACH(pci_data->vm_pci_device_list, vm_pci_device_list, dev) {
		pci_d3d0_reset_devices(dev);
	}

	/* wait 2 sec */
	usleep(2 * 1000 * 1000);

	/* Reinit devices */
	LIST2_FOREACH(pci_data->vm_pci_device_list, vm_pci_device_list, dev) {
		pci_reinit_devices(dev, true);
	}

	/* wait 2 sec */
	usleep(2 * 1000 * 1000);
}

DRIVER_PREINIT(pci_preinit);
BUS_VMINIT(pci_setup_vmresource);
BUS_RESET(pci_reset_assigned_device);
